/*      enet.c  CMU     1/18/80 */

/*
 *  Ethernet interface driver
 *
 **********************************************************************
 * HISTORY
 * 25-Mar-81  Jeffrey Mogul at Stanford
 *	Created separate, smaller free queue for transmitted packets.
 *	This allows a large freeq for received packets, and thus
 *	more ethernet minor devices.  Fixed minor bug in priority
 *	of error returns (ENXIO checked before EBUSY.)
 *
 * 02-Feb-81  Jeffrey Mogul at Stanford
 *	Installed bug-fixes due to Peter Eichenberger:
 *		"Fixed unprotected critical sections which caused
 *		 garbage packets to be sent and received packets
 *		 to be lost.  Fixed bugs in check that UNIBUS
 *		 addresses fall within frst 64K bytes.  Added check
 *		 for illegal minor dev number in open."
 *
 * 01-Nov-80  Mike Accetta (mja) at Carnegie-Mellon University
 *	Upgraded for Berkeley version 3.34;  extracted NENET and
 *	ENADDR defintions onto en.h;  added support for FIONREAD ioctl
 *	call;  removed support for FIOCCHKR ioctl call;  added dummy
 *	reset routine for UNIBUS resets until they become a problem
 *	worth filling in the code for; fixed bug in recording process
 *	pointer for signals (V2.00).
 *
 * 04-Jun-80  Mike Accetta (mja) at Carnegie-Mellon University
 *	Changed to return immediately from read with EOF
 *	if no input packet is available and the timeout value is
 *	less than 0;  also added EIOCENBS and EIOCINHS ioctl calls
 *	to enable and inhibit signals when an input packet is
 *	received (V1.07f).
 *
 * 01-Apr-80  Mike Accetta (mja) at Carnegie-Mellon University
 *	Changed back to using separate BDP's for input and output
 *	packets; alsoo added check that alloacted unibus addresses fall
 *	within the first 64K bytes (V1.06b).
 *
 * 27-Feb-80  Mike Accetta (mja) at Carnegie-Mellon University
 *	Modified enread() to correctly interpret zero length reads from
 *	the FIOCCHKR ioctl call;  added EIOCGETP, EIOCSETP and EIOCSETF
 *	as synonyms for TIOCGETP, TIOCSETP, and TIOCSETD so that the
 *	tty versions may eventually be removed;  added check for minor
 *	device below HIPRIDEV when setting priorities above HIPRI
 *	(V1.05a).
 *
 * 22-Feb-80  Rick Rashid (rfr) at Carnegie-Mellon University
 *	Rewritten to provide multiple user access via user settable
 *	filters (V1.05).
 *
 * 18-Jan-80  Mike Accetta (mja) at Carnegie-Mellon University
 *      Created (V1.00).
 *
 **********************************************************************
 */

#include "../conf/enet.h"

#if NENET > 0

#include "../h/param.h"
#include "../h/systm.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/map.h"
#include "../h/pte.h"
#include "../h/buf.h"
#include "../h/uba.h"
#include "../h/tty.h"
#include "../h/queue.h"
#include "../h/enet.h"

#define DEBUG	0

#define	FIOCCHKR	(('f'<<8)|102)

#define enraddr(addr)   ((caddr←t)(addr) - (caddr←t)(&RQueueElts[0]))+enrbgn
#define enxaddr(addr)   ((caddr←t)(addr) - (caddr←t)(&XQueueElts[0]))+enxbgn

#define ENETBUSY      01
#define ENETTIMEDOUT  02

#define PRINET  26

#define ERROR   0100000
#define IE      0100
#define GO      01

/*
 *  The following definitions are meant to clean up
 *  C language syntax a bit.
 */

#define boolean int
#define integer int
#define true    1
#define false   0
#define endcase break;

/*
 *  Ethernet packet structure.
 */

struct Packet
{
    short Data[PSIZE+2];
};

/*
 *  An OpenDescriptor is the data structure used to keep
 *  state information about a particular open enet file.
 *  In particular, it keeps track of the current timeout
 *  value for that file, the current packet filter, and
 *  is identified by the file pointer used to access it.
 *  There are MAXOPENS possible OpenDescriptors and all
 *  OpenDescriptors corresponding to active files are
 *  kept in a linked list headed by 'endesq'.  When an
 *  input packet arrives, that list is scanned and each
 *  OpenDescriptor's filter is applied to the packet until
 *  a true value is returned. The packet is then queued
 *  into the Waiting list for that OpenDescriptor. (Unless
 *  MAXWAITING packets are already in the queue, at which
 *  point the packet is returned to the free list.)
 *  Note that when a single local variable is needed to
 *  which conforms to this structure, the abbreviated
 *  variable name 'd' will be used.
 */

struct OpenDescriptor
{
    struct Queue      Link;              /* Linked list of OpenDescriptors */
    struct Queue      Waiting;           /* Linked list of waiting packets */
    integer           State;             /* ENETTIMEDOUT or ENETBUSY */
    integer           NumQueued;         /* Number of waiting packets */
    integer           Timeout;           /* Length of time to wait for packet */
    integer	      SigNumb;		 /* Signal number for input packet notification */
    struct proc	     *SigProc;		 /* Process to signal */
    struct enetfilter OpenFilter;        /* Packet filter */
} AllDescriptors[MAXOPENS];

/*
 *  Ethernet packets are embedded in queue elements which
 *  contain a variety of information needed about the packet.
 */

struct QueueElt
{
    struct Queue      Link;
    integer           Residual;
    integer           Count;
    integer           Delay;
    integer           Mask;
    integer           Rand;
    struct Packet     Pkt;
} RQueueElts[NPACK];

struct	QueueElt XQueueElts[NXMTFREE];

/*
 *  'enRfreeq' and 'enXfreeq' are the heads of the receive and transmit
 *  free queues. 'enxmitq' is the head of the queue of packets waiting to be
 *  transmitted.
 */

struct Queue enRfreeq;
struct Queue enXfreeq;
struct Queue enxmitq;
struct Queue endesq;

/*
 *  Various needed global variables.
 */

integer enrmap;  	  /* Unibus mapping for input packets */
integer enxmap;  	  /* Unibus mapping for output packets */
integer enopens;          /* Number of simultaneous enet opens */
integer enopf[MAXOPENS];
integer enrerrs;     	  /* Number of errors on receive */
integer enxerrs;     	  /* Number of errors on transmit */
integer enrbgn;		  
integer enxbgn;		  

/*
 *  'rp' and 'xp' point to the packets currently being transmitted
 *  or received by the ethernet.  If 'rp', for example, is NULL, this
 *  implies no outstanding input request (i.e. the ethernet receiver
 *  is idle). Likewise for 'xp' and the transmitter.
 */

struct QueueElt * rp;
struct QueueElt * xp;

/*
 *  Structure of device register area for ethernet.
 */

struct enreg
{
    short enxwc;
    short enxba;
    short enxcsr;
    short enxdly;
    short enrwc;
    short enrba;
    short enrcsr;
    short enaddr;
};

/*
 *  Forward declarations for subroutines which return other
 *  than integer types.
 */

struct Queue * dequeue();
boolean queueempty();


/*
 *  enopen - open ether net device
 */

enopen(dev, flag)
{
    register integer i;
    register integer md;
    register struct OpenDescriptor * d;

    /*
     * Each open enet file has a different minor device number.
     * In general, a prospective net user must try successively
     * to open the devices 'enet', 'enet1', ... 'enetn' where n
     * is MAXOPENS-1.  This is not elegant, but UNIX will call
     * open for each new open file using the same inode but calls
     * close only when the last open file referring to the inode 
     * is released. This means that we cannot know inside the
     * driver code when the resources associated with a particular
     * open of the same inode should be deallocated.  Thus, we have
     * simply made up a number of different inodes to represent the
     * simultaneous opens of the ethernet.  Each of these has
     * a different minor device number.
     */

    md = minor(dev);

    if (md >= MAXOPENS)		/* check for illegal minor dev first */
        u.u←error = ENXIO;
    else if (enopf[md] /*|| dr←open[0] || dr←open[1]*/)
        u.u←error = EBUSY;
    else
    {
        enopf[md] = true;
        if (enopens == 0)
	{

	    /*
	     * First open.  We must initialize unibus map and 
	     * various queues and data structures of interest.
	     * Since the ethernet interface ignores the extended
	     * memory address bits, we must insure that the
	     * allocated portions of the unibus map are all in
	     * the first 64K before proceeding.
	     */

	    for (;;)
	    {
		enrmap = uballoc(&RQueueElts[0], sizeof RQueueElts, 1);
		enxmap = uballoc(&XQueueElts[0], sizeof XQueueElts, 1);
		/* check that the ENDs of the UNIBUS spaces allocated
		 * are both inside the 64K byte addressing limit of the
		 * interface
		 */
		if ( ((enrmap + sizeof RQueueElts)&0x30000) ||
			((enxmap + sizeof XQueueElts)&0x30000))
		{
		    ubafree(enrmap);
		    ubafree(enxmap);
		    enrmap = enxmap = 0;
		    if (((sizeof RQueueElts)+(sizeof XQueueElts)) >= 0x10000)
			/* oops - we will NEVER find enough UNIBUS space
			 * for all of these packets, get out now */
			 return(ENXIO);
		    sleep((caddr←t)&lbolt, PWAIT);
		}
		else
		    break;
	    }
            enrbgn = enrmap & 0x3ffff;
            enxbgn = enxmap & 0x3ffff;
#if	DEBUG
            printf("\nenopen, enrmap = %x, enrbgn = %x, enxmap = %x, enxbgn = %x",
		   enrmap,enrbgn,enxmap,enxbgn);
#endif
            enrerrs = 0;
            enxerrs = 0;
            xp = NULL;
            rp = NULL;
            initqueue(&enRfreeq);
            initqueue(&enXfreeq);
            initqueue(&enxmitq);
            initqueue(&endesq);
            for (i = 0; i < NPACK; i++)
		enqueue(&enRfreeq,&RQueueElts[i].Link);
            for (i = 0; i < NXMTFREE; i++)
		enqueue(&enXfreeq,&XQueueElts[i].Link);
            enrstart();
	}        
	enopens++;
	eninitdescriptor(md);
    }
}

/*
 *  enclose - ether net device close routine
 */

enclose(dev, flag)
  {
    register struct OpenDescriptor * d;
    register struct QueueElt * p;
    register integer i;

    i = minor(dev);

    enopf[i] = false;      
    enopens--;
    d = &AllDescriptors[i];

    remqueue(&endesq,&(d->Link));

    while (! queueempty(&(d->Waiting)))
      {
        p = (struct QueueElt *)dequeue(&(d->Waiting));
        enqueue(&enRfreeq,&(p->Link));
      }

    if (enopens == 0)
      {
        /*
	 *  All done. No enet files are now open.  We can
         *  close down shop completely.
  	 */
	
        ENADDR->enrcsr = 0;
        ENADDR->enxcsr = 0;
        if (enrmap) ubafree(enrmap);
        if (enxmap) ubafree(enxmap);
        enrmap = 0;
        enxmap = 0;
        xp = NULL;
        rp = NULL;
      }
  }

/*
 *  enread - read next packet from net
 */

enread(dev)
{

    register struct OpenDescriptor * d;
    register struct QueueElt * p;
    extern enrtout(), enrstart();

#if	DEBUG
    printf("\nenread called");
#endif

    spl7();

    /*  
     *  Get descriptor for this open enet file.
     */

    d = &AllDescriptors[minor(dev)];

    /*  If nothing is on the queue of packets waiting for
     *  this open enet file, then set timer and sleep until
     *  either the timeout has occurred or a packet has
     *  arrived.
     */


    while ((p = (struct QueueElt *)dequeue(&(d->Waiting))) == NULL)
    {
	if (d->Timeout < 0)
	{
	    spl0();
	    return;
	}
        if (d->Timeout)
	{
            timeout(enrtout, d, d->Timeout);
            d->State |= ENETBUSY;
	}

        d->State &= ~(ENETTIMEDOUT);
        sleep(d, PRINET);

        if (d->State & ENETBUSY)
	{
            untimeout(enrtout, d);
            d->State &= ~ENETBUSY;
	}

        if (d->State & ENETTIMEDOUT)
	{
            spl0();
            return;
	}
    }

    d->NumQueued--;
    spl0();

    /*  
     *  Move data from packet into user space.  Throw away
     *  any data left over if u.u←count is less than 
     *  the number of 16-bit words in the packet.
     */

    iomove((caddr←t)&(p->Pkt),
            min(((p->Residual)+PSIZE)<<1, u.u←count), 
            B←READ);

    enqueue(&enRfreeq,&(p->Link));
    spl6();
    if (rp == NULL) enrstart();	/* start read if no packet being received */
    spl0();
}

/*
 *  enwrite - write next packet to net
 */

enwrite(dev)
{
    register struct QueueElt * p;
    extern enxstart();

#if	DEBUG
    printf("\nenwrite called");
#endif

    spl6();

    /*  
     *  If there are no free packets for transmission, then
     *  sleep until there are.  
     */

    while ((p = (struct QueueElt *) dequeue(&enXfreeq)) == NULL)
    {
        sleep(&enXfreeq, PRINET);
    }

    spl0();

    iomove((caddr←t)&(p->Pkt),
           p->Count=min(PSIZE<<1, u.u←count), 
           B←WRITE);

    p->Mask = -1;
    p->Rand = time;
    p->Delay = 0;
   
    enqueue(&enxmitq,&(p->Link));
    spl6();
    if (xp == NULL) enxstart();	/* start write if no packet being sent */
    spl0();
}

/*
 *  enioctl - ether net control
 */

enioctl(dev, cmd, addr, flag)
caddr←t addr;
dev←t flag;
{

    integer n;
    struct eniocb t;
    struct enetfilter f;
    struct OpenDescriptor * d;

#if	DEBUG
    printf("\nenioctl called with minor device %d",minor(dev));
#endif
    d = &AllDescriptors[minor(dev)];

    switch (cmd)
    {
	case EIOCGETP:
	case TIOCGETP:
          {
            spl6();
            t.en←rerr = enrerrs;
            t.en←xerr = enxerrs;
            spl0();

            t.en←rtout = d->Timeout;
            t.en←addr = ~(ENADDR->enaddr);
	    t.en←maxfilters = MAXFILTERS;

            if (copyout((caddr←t)&t, addr, sizeof t))
                u.u←error = EFAULT;
          }
        endcase

        case TIOCSETP:
        case EIOCSETP:
          {
            if (copyin(addr, (caddr←t)&t, sizeof t))
              {
                u.u←error = EFAULT;
                return;
              }
            d->Timeout = t.en←rtout;
          }
        endcase

        case TIOCSETD:
        case EIOCSETF:
          {
            if (copyin(addr, (caddr←t)&f, sizeof f))
              {
                u.u←error = EFAULT;
                return;
              }

	    if ((minor(dev) < HIPRIDEV && f.Priority >= HIPRI) ||
		f.FilterLen > MAXFILTERS)
	      {
		u.u←error = EINVAL;
		return;
	      }
            spl6();
            bcopy((caddr←t)&f, (caddr←t)&(d->OpenFilter), sizeof f);
            remqueue(&endesq,&(d->Link));
            eninsert(d);
            spl0();
          }
        endcase

	/*
	 *  Enable signal n on input packet
	 */
	case EIOCENBS:
	{
	    if (copyin(addr, (caddr←t)&n, sizeof n))
	    {
		u.u←error = EFAULT;
		return;
	    }
	    d->SigProc = u.u←procp;
	    d->SigNumb = n;
	}
	endcase

	/*
	 *  Disable signal on input packet
	 */
	case EIOCINHS:
	{
	    d->SigNumb = 0;
	}
	endcase

	/*
	 *  Check for packet waiting
	 */
	case FIONREAD:
	case FIOCCHKR:
	{
	    spl6();
	    if (queueempty(&(d->Waiting)))
		n = 0;
	    else
		n = (((struct QueueElt *)(d->Waiting.F))->Residual+PSIZE)<<1;
	    spl0();
	    if (cmd == FIOCCHKR)
	    {
		if (n == 0)
			u.u←error = ENXIO;
	    }
	    else
            if (copyout((caddr←t)&n, addr, sizeof n))
                u.u←error = EFAULT;
	}
	endcase

        default:
          {
            u.u←error = EINVAL;
          }
        endcase
    }

}

/*
 *  enrtout - process ether net read timeout
 */

enrtout(d)
struct OpenDescriptor * d;
  {
    register integer ipl;
    ipl = spl6();
    d->State |= ENETTIMEDOUT;
    d->State &= ~(ENETBUSY);
    wakeup(d);
    splx(ipl);
  }

/*
 *  enrstart - start read operation on net
 */

enrstart()
  {
    register struct QueueElt * p;

    if (rp == NULL)
      {
        if ((rp = (struct QueueElt *)dequeue(&enRfreeq))== NULL)
          return;
      }

#if	DEBUG
    printf("\nenrstart ENADDR->enrba is %x",enraddr((caddr←t)&(rp->Pkt)));
    printf("\nrp = %x",rp);
#endif
    ENADDR->enrwc = -PSIZE;
    ENADDR->enrba = enraddr((caddr←t)&(rp->Pkt));
    ENADDR->enrcsr = IE|GO;

#if	DEBUG
    printf("\ngot beyond enet start");
#endif

    return;
  }

/*
 * enxstart - start packet transmission on net
 */

enxstart()
  {
    if (xp == NULL)
      {
        if ((xp = (struct QueueElt *)dequeue(&enxmitq)) == NULL)
          return;
      }

#if	DEBUG
    printf("\nenxstart ENADDR->enxba is %x",enxaddr((caddr←t)&(xp->Pkt)));
#endif
    ENADDR->enxwc = -((xp->Count+1)>>1);
    ENADDR->enxba = enxaddr((caddr←t)&(xp->Pkt));
    ENADDR->enxdly = xp->Delay;
    ENADDR->enxcsr = IE|GO;


    return;
  }

/*
 * enrint - net read interrupt handler
 */

enrint()
  {

#if	DEBUG
    printf("\nenrint called");
#endif
    if (rp == NULL) return;

    if (ubapurge(enrmap) || (ENADDR->enrcsr&ERROR))
      {
        if (enrerrs < MAXERR) enrerrs++;
        enrstart();
        return;
      }

    rp->Residual = ENADDR->enrwc;
    if (rp->Residual) rp->Residual |= (-1&~01777);

    eninputdone(rp);

    rp = NULL;
    enrstart();
  }

/*
 * eninputdone - process correctly received packet
 */

eninputdone(p)
struct QueueElt *p;
{
    struct OpenDescriptor * d;
    register integer ipl;

#if	DEBUG
    printf("\neninputdone called");
#endif
    d = (struct OpenDescriptor *) endesq.F;
    while (&endesq != &(d->Link))
    {
        if (enfilter(p,d))
	{
            if (d->NumQueued < MAXWAITING)
	    {
                enqueue(&(d->Waiting),&(p->Link));
                d->NumQueued++;
                wakeup(d);
		/*  send signal when input packet received  */
		if (d->SigNumb)
		{
		    psignal(d->SigProc, d->SigNumb);
		    d->SigNumb = 0;		/* disable signal */
		}
#if	DEBUG
                printf("\neninputdone packet queued");
#endif
                return;
	    }
            else
	    {
                break;
	    }
	}
        d = (struct OpenDescriptor *)d->Link.F;
    }
    enqueue(&enRfreeq,&(p->Link));
    wakeup(&enRfreeq);
#if	DEBUG
    printf("eninputdone packet freed");
#endif
}

/*
 * enxint - net transmit interrupt handler
 */

enxint()
  {
#if	DEBUG
    printf("\nenxint called");
#endif
    if (xp == NULL) return;
     
    if (ubapurge(enxmap) || (ENADDR->enxcsr&ERROR))
      {
        if (xp->Mask == 0)
          {
            if (enxerrs < MAXERR) enxerrs++;
          }
        else
          {
            xp->Mask <<= 1;
            xp->Delay = xp->Rand & ~xp->Mask;
            enxstart();
            return;
          }
      }

    enqueue(&enXfreeq,&(xp->Link));
    wakeup(&enXfreeq);

    xp = NULL;
    enxstart();
  }

eninitdescriptor(i)
integer i;
{
#if	DEBUG
    printf("\neninitdescriptor called");
#endif
    AllDescriptors[i].NumQueued = 0;
    AllDescriptors[i].State = 0;
    AllDescriptors[i].OpenFilter.FilterLen = 0;
    AllDescriptors[i].OpenFilter.Priority = 0;
    AllDescriptors[i].Timeout = 0;
    AllDescriptors[i].SigNumb = 0;
    initqueue(&AllDescriptors[i].Waiting);
    eninsert(&AllDescriptors[i]);
#if	DEBUG
    printf("\neninitdescriptor returns");
#endif
}

#define	opx(i)	(i>>NBPA)

boolean enfilter(p,d)
struct QueueElt * p;
struct OpenDescriptor * d;
{

    register unsigned short *sp;
    register unsigned short *fp;
    register unsigned short *fpe;
    register unsigned op;
    register unsigned arg;
    register maxword;
    unsigned short stack[MAXFILTERS+1];
    struct fw {unsigned arg:NBPA, op:NBPO;};

    sp = &stack[MAXFILTERS];
    maxword = p->Residual+PSIZE+PUSHWORD;
    fp = &d->OpenFilter.Filter[0];
    fpe = &fp[d->OpenFilter.FilterLen];
    *sp = true;

    for (; fp < fpe; )
    {
	op = ((struct fw *)fp)->op;
	arg = ((struct fw *)fp)->arg;
	fp++;
	switch (arg)
	{
	    default:
		if ((arg >= PUSHWORD)&&(arg < maxword))
		    *--sp = p->Pkt.Data[arg-PUSHWORD];
		else
		    return(false);
		break;
	    case PUSHLIT:
		*--sp = *fp++;
		break;
	    case PUSHZERO:
		*--sp = 0;
	    case NOPUSH:
		break;
	}
	if (op == NOP)
	    continue;
	if (sp > &stack[MAXFILTERS-2])
	    return(false);
	arg = *sp++;
	switch (op)
	{
	    default:
		return(false);
	    case opx(AND):
		*sp &= arg;
		break;
	    case opx(OR):
		*sp |= arg;
		break;
	    case opx(XOR):
		*sp ↑= arg;
		break;
	    case opx(EQ):
		*sp = (*sp == arg);
		break;
	    case opx(LT):
		*sp = (*sp < arg);
		break;
	    case opx(LE):
		*sp = (*sp <= arg);
		break;
	    case opx(GT):
		*sp = (*sp > arg);
		break;
	    case opx(GE):
		*sp = (*sp >= arg);
		break;
	}
    }
    return(*sp);

}
	

    
eninsert(d)
struct OpenDescriptor * d;
  {
    struct OpenDescriptor * nxt;
    register integer ipl;
    ipl = spl6();
    nxt = (struct OpenDescriptor *) endesq.F;
    while (&endesq != &(nxt->Link))
      {
        if (d->OpenFilter.Priority > nxt->OpenFilter.Priority) break;
        nxt = (struct OpenDescriptor *)nxt->Link.F;
      }
    enqueue(&(nxt->Link),&(d->Link));
    splx(ipl);
  }

enreset()
{

    panic("enreset");

}
#endif
