/*      enet.h  CMU     1/18/80 */

/*
 *  Ethernet definitions needed for user processes
 *
 **********************************************************************
 * HISTORY
 * 25-Mar-81  Jeff Mogul @ Stanford
 *	In conjunction with changes to the driver (enet.c), modified
 *	to allow more minor devices and more packets waiting per
 *	input device.
 *
 * 16-Mar-81  Jeff Mogul @ Stanford
 *	Changed declaration of en←rtout from signed to unsigned,
 *	to agree with documentation and common sense.
 *
 * 03-Jul-80  Mike Accetta (mja) at Carnegie-Mellon University
 *	Added EIOCENBS and EIOCINHS definitions (V1.07f).
 *
 * 27-Feb-80  Mike Accetta (mja) at Carnegie-Mellon University
 *	Added EIOCGETP, EIOCSETP, EIOCSETF, HIPRIDEV, and HIPRI
 *	definitions (V1.05a).
 *
 * 22-Feb-80  Rick Rashid (rfr) at Carnegie-Mellon University
 *	Rewritten for multiple simultaneous opens with filters (V1.05).
 *
 * 18-Jan-80  Mike Accetta (mja) at Carnegie-Mellon University
 *      Created (V1.00).
 *
 **********************************************************************
 */

/*
 * Since the ethernet interface cannot address more than 64kb,
 * this limits the number of packets that can be queued for I/O.
 * The size of a QueueElt (the mapped unit) is 632 bytes, so
 * we could fit up to 65535/632 == 103 of them.
 */
#define MAXPACKETS 100			/* maximum number of packets
					 * that can be mapped */

#define MAXWAITING  3			/* maximum number of packets
					 * which can be waiting at one
					 * time for an open enet file 
					 */
#define MAXOPENS   31			/* # of enet devices */
#define	HIPRIDEV   12			/* lowest high priority device */
#define	HIPRI	   128			/* high priority priority */
#define MAXERR     255                  /* maximum recordable errors */
#define PSIZE      300                  /* size of packet in short words */
					/* if you change this, change
					 * MAXPACKETS, too ! */
#define NPACK      (MAXWAITING*MAXOPENS)/* number of receiver packets */
#define NXMTFREE   (MAXPACKETS-NPACK)	/* number of transmitter packets */
					/* this should be > ca. 4 - 6 */

#define MAXFILTERS 40			/* maximum filter short words */

/*  status bits in csr  */
#define PROM    02                      /* promiscuous mode (currently unused) */

/*  set/get parameters, set filter ioctl commands  */
#define	EIOCSETP	(('E'<<8)|100)
#define	EIOCGETP	(('E'<<8)|101)
#define	EIOCSETF	(('E'<<8)|102)
#define	EIOCENBS	(('E'<<8)|103)
#define	EIOCINHS	(('E'<<8)|104)

/*
 *  We now allow specification of up to 40 (short) words of a filter
 *  command list to be applied to incoming packets to determine if
 *  those packets should be given to a particular open ethernet file.
 *  
 *  Each open enet file specifies the filter command list via iocontrl.
 *  Each filter command list specifies a sequences of actions which leave a
 *  boolean value on the top of an internal stack.  Each word of the
 *  command list specifies an action from the set {PUSHLIT, PUSHZERO,
 *  PUSHWORD+N} which respectively push the next word of the stack, zero,
 *  or word N of the incoming packet on the stack, and a binary operator
 *  from the set {EQ, LT, LE, GT, GE, AND, OR, XOR} which operates on the
 *  top two elements of the stack and replaces them with its result.  The
 *  special action NOPUSH and the special operator NOP can be used to only
 *  perform the binary operation or to only push a value on the stack.
 *  
 *  If the final value of the filter operation is true, then the packet is
 *  accepted for the open file which specified the filter.  If it is false,
 *  the next open file's filter is examined.
 *  
 *  Open file filters are ordered according to their priority.  Lowest
 *  priority is 0, highest is 255.  Priorities from HIPRI up may  not be
 *  set from ethernet files with minor device below HIPRIDEV.  Open
 *  ethernet files with the same priority are handled in the order they
 *  were opened.
 */

/*  these must sum to 16!  */
#define	NBPA	10			/* # bits / action */
#define	NBPO	6			/* # bits / operator */

/*  binary operators  */
#define NOP	(0<<NBPA)
#define	EQ	(1<<NBPA)
#define	LT	(2<<NBPA)
#define	LE	(3<<NBPA)
#define	GT	(4<<NBPA)
#define GE	(5<<NBPA)
#define	AND	(6<<NBPA)
#define	OR	(7<<NBPA)
#define	XOR	(8<<NBPA)

/*  stack actions  */
#define	NOPUSH	 0
#define	PUSHLIT	 1	
#define	PUSHZERO 2
#define	PUSHWORD 16

/*  parameter buffer structure  */
struct eniocb
{
    char en←addr;                       /* ethernet address (RO) */
    unsigned char en←maxfilters;	/* maximum filter words available (RO) */
    unsigned char en←rerr;              /* receive errors since last GETP (RO) */
    unsigned char en←xerr;              /* transmit errors since last GETP (RO) */
    unsigned short en←rtout;            /* receive timeout in (jiffies) (RW) */
    short en←flags;                     /* status bits (csr) (currently unused) */
};

/*
 *  filter structure for SETF
 */
struct enetfilter
{
    unsigned char Priority;		/* priority of filter */
    unsigned char FilterLen;		/* length of filter command list */
    unsigned short Filter[MAXFILTERS];	/* the filter command list */
};
